
import logging
import random
import os
import time

from django.contrib.auth.backends import ModelBackend
from django.http import Http404
from django.utils import timezone

import ratelimit

from .models import UserComponent, Protection, TravelProtection, AuthToken
from .constants import (
    ProtectionType, MIN_PROTECTION_STRENGTH_LOGIN,
)

logger = logging.getLogger(__name__)

# seed with real random
_nonexhaustRandom = random.Random(os.urandom(30))


class SpiderTokenAuthBackend(ModelBackend):

    def authenticate(self, request, username=None,
                     protection_codes=None, **kwargs):
        """ Use protections for authentication"""
        tokenstring = request.GET.get("token", None)
        if not tokenstring:
            return

        now = timezone.now()

        token = AuthToken.objects.filter(
            usercomponent__name="index",
            token=tokenstring
        ).first()
        if token:
            uc = token.usercomponent
            if token.persist == -1 and token.created < now - uc.token_duration:
                token.delete()
                return
            if TravelProtection.objects.auth(request, uc, now):
                return uc.user


class SpiderAuthBackend(ModelBackend):

    def authenticate(self, request, username=None,
                     protection_codes=None, nospider=False, **kwargs):
        """ Use protections for authentication"""
        # disable SpiderAuthBackend backend (against recursion)
        if nospider:
            return
        uc = UserComponent.objects.filter(
            user__username=username, name="index"
        ).first()
        if not uc:
            request.protections = Protection.authall(
                request, scope="auth",
                ptype=ProtectionType.authentication.value,
                protection_codes=protection_codes
            )
            if type(request.protections) is int:  # should never happen
                logger.warning(
                    "Login try without username, should never "
                    "happen, archieved strength: %s",
                    request.protections
                )
                return None
        else:
            try:
                request.protections = uc.auth(
                    request, scope="auth",
                    ptype=ProtectionType.authentication.value,
                    protection_codes=protection_codes
                )
            except Http404:
                # for Http404 auth abort by protections (e.g. Random Fail)
                pass

            if type(request.protections) is int:
                if TravelProtection.objects.auth(request, uc):
                    if request.protections < MIN_PROTECTION_STRENGTH_LOGIN:
                        logger.warning(
                            "Low login protection strength: %s, %s",
                            request.protections, username
                        )
                    return uc.user
        # error path

        # allow blocking per hour
        ratelimit.get_ratelimit(
            request=request, group="spider_login_failed_ip", key="ip",
            inc=True, rate=(float("inf"), 3600)
        )
        ratelimit.get_ratelimit(
            request=request, group="spider_login_failed_account",
            key=lambda x, y: username, inc=True, rate=(float("inf"), 3600)
        )
        # be less secure here, most probably the user is already known
        time.sleep(_nonexhaustRandom.random()/2)



# Generated by Django 2.1.5 on 2019-01-28 22:30

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0004_auto_20190128_1349'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authtoken',
            name='token',
            field=models.CharField(db_index=True, max_length=138, unique=True, validators=[django.core.validators.RegexValidator('^[-a-zA-Z0-9_/]+\\Z', 'Enter a valid token.', 'invalid')]),
        ),
    ]



# Generated by Django 2.1.5 on 2019-02-11 23:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0006_auto_20190128_2251'),
    ]

    operations = [
        migrations.AddField(
            model_name='usercomponent',
            name='allow_domain_mode',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-31 12:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0010_assignedcontent_attached_to_primary_anchor'),
    ]

    operations = [
        migrations.AddField(
            model_name='usercomponent',
            name='default_content_features',
            field=models.ManyToManyField(blank=True, help_text='Select features used by default for contents in this component', limit_choices_to=models.Q(ctype__contains='g'), related_name='default_feature_for_contents', to='spider_base.ContentVariant'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='primary_anchor',
            field=models.ForeignKey(blank=True, help_text='Select main identifying anchor. Also used for attaching persisting tokens (elsewise they are attached to component)', limit_choices_to={'info__contains': '\nanchor\n'}, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='primary_anchor_for', to='spider_base.AssignedContent'),
        ),
    ]



__all__ = ["SpiderBaseConfig"]

from django.apps import AppConfig
from django.db.models.signals import (
    post_migrate, post_save, pre_save, post_delete, m2m_changed
)
from django.contrib.auth import get_user_model
from django.conf import settings
from .helpers import extract_app_dicts
from .signals import (
    UpdateSpiderCb, InitUserCb, UpdateAnchorComponentCb, UpdateContentCb,
    update_dynamic, TriggerUpdate, CleanupCb, DeleteContentCb,
    UpdateComponentFeaturesCb, UpdateContentFeaturesCb
)


class SpiderBaseConfig(AppConfig):
    name = 'spkcspider.apps.spider'
    label = 'spider_base'
    spider_url_path = 'spider/'
    verbose_name = 'spkcspider base'

    def ready(self):
        from .models import (
            AssignedContent, UserComponent, LinkContent
        )
        from django.apps import apps
        from .protections import installed_protections

        for app in apps.get_app_configs():
            installed_protections.update(
                extract_app_dicts(app, "spider_protections", "name")
            )

        pre_save.connect(
            UpdateAnchorComponentCb, sender=UserComponent,
        )
        post_save.connect(
            UpdateContentCb, sender=AssignedContent,
        )
        m2m_changed.connect(
            UpdateComponentFeaturesCb, sender=UserComponent.features.through
        )
        m2m_changed.connect(
            UpdateContentFeaturesCb, sender=AssignedContent.features.through,
        )

        # order important for next two
        post_delete.connect(
            CleanupCb, sender=UserComponent,
        )

        post_delete.connect(
            CleanupCb, sender=AssignedContent,
        )

        post_delete.connect(
            DeleteContentCb, sender=LinkContent,
        )

        #####################

        post_save.connect(
            InitUserCb, sender=get_user_model(),
        )

        update_dynamic.connect(
            UpdateSpiderCb,
            dispatch_uid="update_spider_dynamic"
        )

        if getattr(settings, "UPDATE_DYNAMIC_AFTER_MIGRATION", True):
            post_migrate.connect(
                TriggerUpdate, sender=self,
            )



# Generated by Django 2.2.1 on 2019-05-08 20:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0018_auto_20190505_1336'),
    ]

    operations = [
        migrations.AddField(
            model_name='contentvariant',
            name='valid_feature_for',
            field=models.ManyToManyField(blank=True, related_name='valid_features', to='spider_base.ContentVariant'),
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-22 19:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0009_auto_20190317_1405'),
    ]

    operations = [
        migrations.AddField(
            model_name='assignedcontent',
            name='attached_to_primary_anchor',
            field=models.BooleanField(default=False, editable=False, help_text='Content references primary anchor'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-01-28 13:39

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0002_20190127_squashed'),
    ]

    operations = [
        migrations.AddField(
            model_name='assignedcontent',
            name='token',
            field=models.CharField(blank=True, db_index=True, max_length=138, null=True, unique=True, validators=[django.core.validators.RegexValidator('^[-a-zA-Z0-9_/]+\\Z', 'Enter a valid token.', 'invalid')]),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='token',
            field=models.CharField(blank=True, db_index=True, max_length=138, null=True, unique=True, validators=[django.core.validators.RegexValidator('^[-a-zA-Z0-9_/]+\\Z', 'Enter a valid token.', 'invalid')]),
        ),
        migrations.AlterField(
            model_name='assignedcontent',
            name='nonce',
            field=models.SlugField(blank=True, db_index=False, max_length=120, null=True),
        ),
        migrations.AlterField(
            model_name='authtoken',
            name='token',
            field=models.CharField(db_index=True, max_length=137, unique=True, validators=[django.core.validators.RegexValidator('^[-a-zA-Z0-9_/]+\\Z', 'Enter a valid token.', 'invalid')]),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='nonce',
            field=models.SlugField(blank=True, db_index=False, max_length=120, null=True),
        ),
    ]



# Generated by Django 2.1 on 2018-08-10 14:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='textfilet',
            name='non_public_edit',
            field=models.BooleanField(default=False, help_text='Allow others to edit text file if not public'),
        ),
    ]



__all__ = ["UserComponentForm", "UserContentForm"]

import logging
from statistics import mean

from django.conf import settings
from django import forms
from django.db import models
from django.core.exceptions import NON_FIELD_ERRORS
from django.utils.translation import gettext_lazy as _

from ..models import (
    Protection, UserComponent, AssignedContent, ContentVariant, AuthToken,
    TravelProtection
)

from ..helpers import create_b64_id_token
from ..constants import (
    ProtectionType, VariantType, protected_names, MIN_PROTECTION_STRENGTH_LOGIN
)
from ..conf import STATIC_TOKEN_CHOICES, INITIAL_STATIC_TOKEN_SIZE
# from ..widgets import Select2Widget

logger = logging.getLogger(__name__)

_help_text_static_token = _("""Generate a new static token with variable strength<br/>
Tokens protect against bruteforce and attackers<br/>
If you have problems with attackers (because they know the token),
you can invalidate it with this option and/or add protections<br/>
<table style="color:red;">
<tr>
<td>Warning:</td><td>this removes also access for all services you gave the link<td/>
</tr>
<tr>
<td style="vertical-align: top">Warning:</td><td>public user components disclose static token, regenerate after removing
public attribute from component for restoring protection</td>
</tr>
</table>
""")  # noqa: E501


_help_text_features_components = _("""
Features which should be active on component and user content<br/>
Note: persistent Features (=features which use a persistent token) require and enable "Persistence"
""")  # noqa: E501

_help_text_features_contents = _("""
Features which should be active on user content<br/>
Note: persistent Features (=features which use a persistent token) require the active "Persistence" Feature on usercomponent. They are elsewise excluded.
""")  # noqa: E501


class UserComponentForm(forms.ModelForm):
    protections = None
    new_static_token = forms.ChoiceField(
        label=_("New static token"), help_text=_help_text_static_token,
        required=False, initial="", choices=STATIC_TOKEN_CHOICES
    )

    master_pw = forms.CharField(
        label=_("Master Password"),
        widget=forms.PasswordInput(render_value=True), required=False,
        help_text=_(
            "If you don't set a password the unlocking may fail "
            "in future (change of server secrets)"
        )
    )

    class Meta:
        model = UserComponent
        fields = [
            'name', 'description', 'public', 'featured',
            'features', 'default_content_features',
            'primary_anchor', 'required_passes', 'token_duration'
        ]
        error_messages = {
            NON_FIELD_ERRORS: {
                'unique_together': _('Name of User Component already exists')
            }
        }
        help_texts = {
            'features': _help_text_features_components,
        }
        widgets = {
            'features': forms.CheckboxSelectMultiple(),
            'default_content_features': forms.CheckboxSelectMultiple(),
            'description': forms.Textarea(
                attrs={
                    "maxlength": settings.SPIDER_MAX_DESCRIPTION_LENGTH+1
                }
            ),
            # 'primary_anchor': Select2Widget(
            #    allow_multiple_selected=False,
            #    attrs={
            #        "style": "min-width: 150px; width:100%"
            #    }
            # ),

        }

    def __init__(self, request, data=None, files=None, auto_id='id_%s',
                 prefix=None, *args, **kwargs):
        super().__init__(
            *args, data=data, files=files, auto_id=auto_id,
            prefix=prefix, **kwargs
        )
        self.fields["new_static_token"].choices = map(
            lambda c: (c[0], c[1].format(c[0])),
            self.fields["new_static_token"].choices
        )
        if not (
            request.user.is_superuser or
            request.user.has_perm('spider_base.can_feature')
        ) or request.session.get("is_travel_protected", False):
            self.fields['featured'].disabled = True

        self.fields["features"].queryset = (
            self.fields["features"].queryset.exclude(
                models.Q(ctype__contains=VariantType.feature_connect.value) &
                ~models.Q(ctype__contains=VariantType.unique.value)
            ).exclude(
                models.Q(name="DomainMode") |
                models.Q(name="DefaultActions")
            ) &
            request.user.spider_info.allowed_content.all()
        ).order_by("name")

        self.fields["default_content_features"].queryset = (
            self.fields["default_content_features"].queryset.exclude(
                name="DomainMode"
            ) &
            request.user.spider_info.allowed_content.all()
        ).order_by("name")

        if self.instance.id:
            # data and files field maybe not filled (to load from component)
            # remove pw if index because it is used for legitimating changes
            if not self.instance.is_index:
                self.initial["master_pw"] = \
                    self.fields["master_pw"].widget.value_from_datadict(
                        request.POST, request.FILES,
                        self.add_prefix("master_pw")
                    )
            if self.instance.name in protected_names:
                self.fields["name"].disabled = True
            if not self.instance.is_public_allowed:
                self.fields["public"].disabled = True
                self.fields.pop("featured", None)
            self.fields["primary_anchor"].queryset = \
                self.fields["primary_anchor"].queryset.filter(
                    usercomponent=self.instance
                )
            if self.instance.is_index:
                self.fields["master_pw"].required = True
                self.fields["master_pw"].label = _(
                    "Master Login Password"
                )
                self.fields["master_pw"].help_text = _(
                    "Enter Password used for user account<br/>"
                    "Note: to migrate data, first enter old password then the "
                    "new one"
                )
                self.fields.pop("features", None)
                self.fields.pop("featured", None)
                self.fields["required_passes"].help_text = _(
                    "How many protections must be passed to login?<br/>"
                    "Minimum is 1, no matter what selected"
                )
                ptype = ProtectionType.authentication.value
            else:

                ptype = ProtectionType.access_control.value
            self.protections = Protection.get_forms(
                data=data, files=files, prefix=prefix, uc=self.instance,
                ptype=ptype, request=request, form=self
            )
            self.protections = list(self.protections)
        else:
            self.fields.pop("primary_anchor", None)
            self.fields["new_static_token"].initial = INITIAL_STATIC_TOKEN_SIZE
            self.fields["new_static_token"].choices = \
                self.fields["new_static_token"].choices[1:]
            self.protections = []
            self.fields["required_passes"].initial = 1

    @property
    def media(self):
        """Return all media required to render the widgets on this form."""
        media = super().media
        for protection in self.protections:
            media = media + protection.media
        return media

    def is_valid(self):
        isvalid = super().is_valid()
        for protection in self.protections:
            if not protection.is_valid():
                isvalid = False
        return isvalid

    def clean_description(self):
        value = self.cleaned_data.get("description", "").strip()
        if len(value) > settings.SPIDER_MAX_DESCRIPTION_LENGTH:
            return "{}…".format(
                value[:settings.SPIDER_MAX_DESCRIPTION_LENGTH]
            )
        return value

    def calc_protection_strength(self):
        prot_strength = None
        max_prot_strength = 0
        if self.instance.id:
            # instant fail strength
            fail_strength = 0
            # regular protections strength
            amount_regular = 0
            strengths = []
            for protection in self.protections:
                if not protection.cleaned_data.get("active", False):
                    continue
                if not protection.is_valid():
                    continue

                if len(str(protection.cleaned_data)) > 100000:
                    raise forms.ValidationError(
                        _("Protection >100 kb: %(name)s"),
                        params={"name": protection}
                    )
                if protection.cleaned_data.get("instant_fail", False):
                    s = protection.get_strength()
                    if s[1] > max_prot_strength:
                        max_prot_strength = s[1]
                    fail_strength = max(
                        fail_strength, s[0]
                    )
                else:
                    s = protection.get_strength()
                    if s[1] > max_prot_strength:
                        max_prot_strength = s[1]
                    strengths.append(s[0])
                    amount_regular += 1
            strengths.sort()
            if self.cleaned_data["required_passes"] > 0:
                if amount_regular == 0:
                    # login or token only
                    # not 10 because 10 is also used for uniqueness
                    strengths = 4
                else:
                    # avg strength but maximal 3,
                    # (4 can appear because of component auth)
                    # clamp
                    strengths = min(round(mean(
                        strengths[:self.cleaned_data["required_passes"]]
                    )), 3)
            else:
                strengths = 0
            prot_strength = max(strengths, fail_strength)

        return (prot_strength, max_prot_strength)

    def clean(self):
        ret = super().clean()
        assert(self.instance.user)
        self.cleaned_data["can_auth"] = False
        self.initial["master_pw"] = self.cleaned_data.pop("master_pw", None)
        if 'name' not in self.cleaned_data:
            # ValidationError so return, calculations won't work here
            return

        if not self.instance.id:
            quota = self.instance.get_component_quota()
            if quota and UserComponent.objects.filter(
                user=self.instance.user
            ).count() >= quota:
                raise forms.ValidationError(
                    _("Maximal amount of usercomponents reached %(amount)s."),
                    code='quota_exceeded',
                    params={'amount': quota},
                )
            # TODO: cleanup into protected_names/forbidden_names
            if self.cleaned_data['name'] in protected_names:
                raise forms.ValidationError(
                    _('Forbidden Name'),
                    code="forbidden_name"
                )

        for protection in self.protections:
            protection.full_clean()
        prot_strength, max_prot_strength = self.calc_protection_strength()

        self.cleaned_data["strength"] = 0
        if self.cleaned_data["name"] == "index":
            if prot_strength < MIN_PROTECTION_STRENGTH_LOGIN:
                self.add_error(None, forms.ValidationError(
                    _("Login requires higher strength: %(strength)s"),
                    code="insufficient_strength_login",
                    params={"strength": MIN_PROTECTION_STRENGTH_LOGIN}
                ))
            self.cleaned_data["strength"] = 10
            return ret
        if not self.cleaned_data["public"]:
            self.cleaned_data["strength"] += 5
            self.cleaned_data["strength"] += prot_strength or 0
        else:
            # check on creation
            if self.cleaned_data["required_passes"] > 0:
                self.cleaned_data["strength"] += 4
        self.cleaned_data["can_auth"] = max_prot_strength >= 4

        min_strength = self.cleaned_data["features"].filter(
            strength__gt=self.cleaned_data["strength"]
        ).aggregate(m=models.Max("strength"))["m"]
        # min_strength is None if no features require a higher strength
        # than provided
        if min_strength:
            self.add_error("features", forms.ValidationError(
                _(
                    "selected features require "
                    "higher protection strength: %(strength)s"
                ),
                code="insufficient_strength",
                params={"strength": min_strength}
            ))
        return self.cleaned_data

    def _save_protections(self):
        for protection in self.protections:
            if not protection.is_valid():
                continue
            protection.save()

    def _save_m2m(self):
        super()._save_m2m()
        self._save_protections()
        if self.cleaned_data["new_static_token"] != "":
            if self.instance.token:
                print(
                    "Old nonce for Component id:", self.instance.id,
                    "is", self.instance.token
                )
            self.instance.token = create_b64_id_token(
                self.instance.id, "_",
                int(self.cleaned_data["new_static_token"])
            )
            self.instance.save(update_fields=["token"])

    def save(self, commit=True):
        self.instance.strength = self.cleaned_data["strength"]
        self.instance.can_auth = self.cleaned_data["can_auth"]
        return super().save(commit=commit)
    save.alters_data = True


class UserContentForm(forms.ModelForm):
    prefix = "content_control"
    new_static_token = forms.ChoiceField(
        label=_("New static token"), help_text=_help_text_static_token,
        required=False, initial="", choices=STATIC_TOKEN_CHOICES
    )
    migrate_primary_anchor = forms.BooleanField(
        label=_("Migrate primary anchor"),
        help_text=_(
            "Set as primary anchor on target component. "
            "Move tokens and all persistent dependencies. "
            "Elsewise anchor persistent tokens on the component and leave the "
            "dependencies untouched."
        ),
        required=False, initial=False
    )

    class Meta:
        model = AssignedContent
        fields = ['usercomponent', 'features', 'name', 'description']
        error_messages = {
            NON_FIELD_ERRORS: {
                'unique_together': _(
                    "Unique Content already exists"
                )
            }
        }
        help_texts = {
            'features': _help_text_features_contents,
        }
        widgets = {
            # 'usercomponent': Select2Widget(
            #     allow_multiple_selected=False,
            #     attrs={
            #         "style": "min-width: 150px; width:100%"
            #     }
            # ),
            'features': forms.CheckboxSelectMultiple(),
            'description': forms.Textarea(
                attrs={
                    "maxlength": settings.SPIDER_MAX_DESCRIPTION_LENGTH+1
                }
            )

        }

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.instance.content.expose_name:
            self.fields["name"].widget = forms.HiddenInput()
        elif self.instance.content.expose_name == "force":
            self.fields["name"].required = True
        if not self.instance.content.expose_description:
            self.fields["description"].widget = forms.HiddenInput()

        self.fields["new_static_token"].choices = map(
            lambda c: (c[0], c[1].format(c[0])),
            self.fields["new_static_token"].choices
        )
        user = self.instance.usercomponent.user

        travel = TravelProtection.objects.get_active_for_request(request)
        query = UserComponent.objects.filter(
            user=user, strength__gte=self.instance.ctype.strength
        ).exclude(travel_protected__in=travel)
        self.fields["usercomponent"].queryset = query

        if VariantType.feature_connect.value in self.instance.ctype.ctype:
            self.fields["usercomponent"].disabled = True
            self.fields["usercomponent"].help_text = _(
                "Features cannot move between components"
            )

        if self.instance.content.force_token_size:
            sforced = [
                ("", ""),
                (
                    str(self.instance.content.force_token_size),
                    _("New ({})").format(
                        self.instance.content.force_token_size
                    )
                )
            ]

        show_primary_anchor_mig = False
        if self.instance.id:
            if (
                request.user == user and
                self.instance.primary_anchor_for.exists()
            ):
                # can only migrate if a) created, b) real user
                show_primary_anchor_mig = True
            if self.instance.content.force_token_size:
                self.fields["new_static_token"].choices = sforced
        else:
            if self.instance.content.force_token_size:
                self.fields["new_static_token"].initial = \
                    self.instance.content.force_token_size
                self.fields["new_static_token"].choices = sforced[1:]
            else:
                self.fields["new_static_token"].initial = \
                    INITIAL_STATIC_TOKEN_SIZE
                self.fields["new_static_token"].choices = \
                    self.fields["new_static_token"].choices[1:]

        if self.instance.usercomponent.is_index:
            # contents in index has no features as they could allow
            # remote access
            self.fields.pop("features", None)
        else:
            if not self.instance.usercomponent.features.filter(
                name="Persistence"
            ):
                self.fields["features"].queryset = \
                    self.fields["features"].queryset.exclude(
                        ctype__contains=VariantType.persist.value
                    )
            user = request.user
            if not user.is_authenticated:
                user = self.instance.usercomponent.user
            self.fields["features"].queryset = (
                self.fields["features"].queryset &
                user.spider_info.allowed_content.all() &
                self.instance.ctype.valid_features.all()
            ).exclude(name="DomainMode").order_by("name")
            self.fields["features"].initial = (
                self.instance.usercomponent.default_content_features.all() &
                self.instance.ctype.valid_features.all()
            )

        if request.user != user and not request.is_staff:
            del self.fields["usercomponent"]

        if not show_primary_anchor_mig:
            del self.fields["migrate_primary_anchor"]

    def clean_description(self):
        value = self.cleaned_data.get("description", "")
        if not self.instance.content.expose_description:
            return value
        value = value.strip()
        if len(value) > settings.SPIDER_MAX_DESCRIPTION_LENGTH:
            return "{}…".format(
                value[:settings.SPIDER_MAX_DESCRIPTION_LENGTH]
            )
        return value

    def clean(self):
        super().clean()
        if "features" in self.cleaned_data:
            min_strength = self.cleaned_data["features"].filter(
                strength__gt=self.instance.usercomponent.strength
            ).aggregate(m=models.Max("strength"))["m"]
            # min_strength is None if no features require a higher strength
            # than provided
            if min_strength:
                self.add_error("features", forms.ValidationError(
                    _(
                        "selected features require "
                        "higher protection strength: %s"
                    ) % min_strength,
                    code="insufficient_strength"
                ))
            if self.cleaned_data["features"].filter(
                ctype__contains=VariantType.domain_mode.value
            ).exists():
                # fixes strange union bug
                self.cleaned_data["features"] = ContentVariant.objects.filter(
                    models.Q(name="DomainMode") |
                    models.Q(
                        id__in=self.cleaned_data["features"].values_list(
                            "id", flat=True
                        )
                    )
                )
        return self.cleaned_data

    def update_anchor(self):
        if not self.instance.primary_anchor_for.exists():
            return
        if self.instance.primary_anchor_for == self.instance.usercomponent:
            return
        if self.cleaned_data.get("migrate_primary_anchor", False):
            # no signals
            self.instance.primary_anchor_for.set(
                self.instance.usercomponent, bulk=True
            )
            tokens = AuthToken.objects.filter(
                persist=self.instance.id
            )
            tokens.update(
                usercomponent=self.instance.usercomponent,
            )

            AssignedContent.objects.filter(
                attached_to_token__in=tokens
            ).update(usercomponent=self.instance.usercomponent)
        else:
            self.instance.primary_anchor_for.clear(bulk=False)
            # token migrate to component via signal

    def _save_m2m(self):
        super()._save_m2m()
        self.update_anchor()
        update_fields = set()
        if not self.instance.name:
            update_fields.add("name")
        if self.instance.token_generate_new_size is not None:
            if self.instance.token:
                print(
                    "Old nonce for Content id:", self.instance.id,
                    "is", self.instance.token
                )
            self.instance.token = create_b64_id_token(
                self.instance.id,
                "_",
                self.instance.token_generate_new_size
            )
            self.instance.token_generate_new_size = None
            update_fields.add("token")
        if update_fields:
            self.instance.save(update_fields=update_fields)

    def save(self, commit=True):
        if self.cleaned_data.get("new_static_token", ""):
            self.instance.token_generate_new_size = \
                int(self.cleaned_data["new_static_token"])
        elif not self.instance.token:
            self.instance.token_generate_new_size = \
                self.fields["new_static_token"].initial
        return super().save(commit=commit)
    save.alters_data = True



__all__ = ["SpiderTagsConfig"]

from django.apps import AppConfig, apps
from .signals import UpdateDefaultLayouts
from spkcspider.apps.spider.helpers import extract_app_dicts
from spkcspider.apps.spider.signals import update_dynamic


class SpiderTagsConfig(AppConfig):
    name = 'spkcspider.apps.spider_tags'
    label = 'spider_tags'
    spider_url_path = 'spidertags/'
    verbose_name = 'spkcspider tags'

    def ready(self):
        from .fields import installed_fields
        for app in apps.get_app_configs():
            installed_fields.update(
                extract_app_dicts(app, "spider_tag_fields")
            )

        update_dynamic.connect(
            UpdateDefaultLayouts,
            dispatch_uid="update_default_layouts"
        )



__all__ = ("content_name_validator", "validator_token")


from django.core.exceptions import ValidationError
from django.core import validators

from django.utils.translation import gettext_lazy as _


def content_name_validator(value):
    if not value.isprintable():
        raise ValidationError(
            _("Contains control characters"),
            code="control_characters"
        )
    if value:
        if value[0].isspace() or value[-1].isspace():
            raise ValidationError(
                _("Contains hidden spaces"),
                code="hidden_spaces"
            )


validator_token = validators.RegexValidator(
    r'\A[-a-zA-Z0-9_]+\Z',
    _("Enter a valid token."),
    'invalid_token'
)



# Generated by Django 2.1.2 on 2018-10-16 20:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0006_textfilet_preview_words'),
    ]

    operations = [
        migrations.AlterField(
            model_name='textfilet',
            name='preview_words',
            field=models.PositiveIntegerField(default=0, help_text='How many words from start should be used for search, seo, search machine preview? (tags are stripped)'),
        ),
    ]



# Generated by Django 2.2.1 on 2019-05-14 18:53

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0023_auto_20190513_1506'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='travelprotection',
            options={'default_permissions': (), 'permissions': [('approve_travelprotection', 'Can approve dangerous TravelProtections')]},
        ),
    ]



# Generated by Django 2.1 on 2018-08-15 11:28

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('spider_tags', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taglayout',
            name='default_verifiers',
            field=jsonfield.fields.JSONField(blank=True, default=list),
        ),
    ]



from django.contrib import admin

# Register your models here.

from .models import (
    TagLayout
)
from .forms import TagLayoutAdminForm


@admin.register(TagLayout)
class TagLayoutAdmin(admin.ModelAdmin):
    form = TagLayoutAdminForm



__all__ = {
    "validate", "valid_wait_states", "verify_download_size",
    "async_validate", "verify", "async_verify"
}

import logging
import binascii
import tempfile
import os

from django.utils.translation import gettext as _
from django.core.files import File
from django.conf import settings
from django.core import exceptions
from django.test import Client

from rdflib import Graph, URIRef, Literal
from rdflib.namespace import XSD
import requests

from spkcspider import celery_app

from spkcspider.apps.spider.constants import spkcgraph
from spkcspider.apps.spider.helpers import merge_get_url, get_settings_func

from .constants import BUFFER_SIZE
from .functions import get_hashob, get_requests_params
from .models import VerifySourceObject, DataVerificationTag

hashable_predicates = set([spkcgraph["name"], spkcgraph["value"]])

valid_wait_states = {
    "RETRIEVING", "HASHING", "STARTED"
}


def hash_entry(triple):
    h = get_hashob()
    if triple[2].datatype == XSD.base64Binary:
        h.update(triple[2].datatype.encode("utf8"))
        h.update(triple[2].toPython())
    else:
        if triple[2].datatype:
            h.update(triple[2].datatype.encode("utf8"))
        else:
            h.update(XSD.string.encode("utf8"))
        h.update(triple[2].encode("utf8"))
    return h.finalize()


def yield_hashes(graph, hashable_nodes):
    for t in graph.triples((None, spkcgraph["value"], None)):
        if (
            t[0] in hashable_nodes and
            t[2].datatype != spkcgraph["hashableURI"]
        ):
            yield hash_entry(t)


def yield_hashable_urls(graph, hashable_nodes):
    for t in graph.triples(
        (None, spkcgraph["value"], spkcgraph["hashableURI"])
    ):
        if t[0] in hashable_nodes:
            yield t


def verify_download_size(length, current_size=0):
    if not length or not length.isdigit():
        return False
    length = int(length)
    if settings.VERIFIER_MAX_SIZE_ACCEPTED < length:
        return False
    return True


def validate(ob, hostpart, task=None):
    dvfile = None
    source = None
    session = requests.session()
    if isinstance(ob, tuple):
        dvfile = open(ob[0], "r+b")
        current_size = ob[1]
    else:
        current_size = 0
        dvfile = tempfile.NamedTemporaryFile(delete=False)

        source = VerifySourceObject.objects.get(
            id=ob
        )
        url = source.get_url()
        params, can_inline = get_requests_params(url)
        if can_inline:
            resp = Client().get(url)
            if resp.status_code != 200:
                session.close()
                dvfile.close()
                os.unlink(dvfile.name)
                raise exceptions.ValidationError(
                    _("Retrieval failed: %(reason)s"),
                    params={"reason": resp.reason},
                    code="error_code:{}".format(resp.status_code)
                )

            c_length = resp.get("content-length", None)
            if not verify_download_size(c_length, current_size):
                resp.close()
                session.close()
                dvfile.close()
                os.unlink(dvfile.name)
                raise exceptions.ValidationError(
                    _("Content too big: %(size)s"),
                    params={"size": c_length},
                    code="invalid_size"
                )
            c_length = int(c_length)
            current_size += c_length
            # clear file
            dvfile.truncate(c_length)
            dvfile.seek(0, 0)

            for chunk in resp:
                dvfile.write(chunk)
            dvfile.seek(0, 0)
            resp.close()
        else:
            try:
                with session.get(
                    url, stream=True, **params
                ) as resp:
                    if resp.status_code != 200:
                        session.close()
                        dvfile.close()
                        os.unlink(dvfile.name)
                        raise exceptions.ValidationError(
                            _("Retrieval failed: %(reason)s"),
                            params={"reason": resp.reason},
                            code="error_code:{}".format(resp.status_code)
                        )

                    c_length = resp.headers.get("content-length", None)
                    if not verify_download_size(c_length, current_size):
                        session.close()
                        dvfile.close()
                        os.unlink(dvfile.name)
                        raise exceptions.ValidationError(
                            _("Content too big: %(size)s"),
                            params={"size": c_length},
                            code="invalid_size"
                        )
                    c_length = int(c_length)
                    current_size += c_length
                    # preallocate file
                    dvfile.truncate(c_length)
                    dvfile.seek(0, 0)

                    for chunk in resp.iter_content(BUFFER_SIZE):
                        dvfile.write(chunk)
                    dvfile.seek(0, 0)
            except requests.exceptions.Timeout:
                session.close()
                raise exceptions.ValidationError(
                    _('url timed out: %(url)s'),
                    params={"url": url},
                    code="timeout_url"
                )
            except requests.exceptions.ConnectionError:
                session.close()
                raise exceptions.ValidationError(
                    _('invalid url: %(url)s'),
                    params={"url": url},
                    code="invalid_url"
                )
    g = Graph()
    g.namespace_manager.bind("spkc", spkcgraph, replace=True)
    try:
        g.parse(
            dvfile.name,
            format="turtle"
        )
    except Exception:
        if settings.DEBUG:
            dvfile.seek(0, 0)
            logging.exception(dvfile.read())
        logging.error("Parsing file failed")
        session.close()
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph fromat'),
            code="invalid_format"
        )

    tmp = list(g.triples((None, spkcgraph["scope"], None)))
    if len(tmp) != 1:
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph'),
            code="invalid_graph"
        )
    start = tmp[0][0]
    scope = tmp[0][2].toPython()
    tmp = list(g.triples((start, spkcgraph["pages.num_pages"], None)))
    if len(tmp) != 1:
        session.close()
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph'),
            code="invalid_graph"
        )
    pages = tmp[0][2].toPython()
    tmp = list(g.triples((
        start,
        spkcgraph["pages.current_page"],
        Literal(1, datatype=XSD.positiveInteger)
    )))
    if len(tmp) != 1:
        session.close()
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Must be page 1'),
            code="invalid_page"
        )
    if task:
        task.update_state(
            state='RETRIEVING',
            meta={
                'page': 1,
                'num_pages': pages
            }
        )
    tmp = list(g.objects(start, spkcgraph["action:view"]))
    if len(tmp) != 1:
        session.close()
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph'),
            code="invalid_graph"
        )
    view_url = tmp[0].toPython()
    if isinstance(ob, tuple):
        split = view_url.split("?", 1)
        source = VerifySourceObject.objects.update_or_create(
            url=split[0], defaults={"get_params": split[1]}
        )
    mtype = set()
    if scope == "list":
        mtype.add("UserComponent")
    else:
        mtype.update(map(
            lambda x: x.toPython(), g.objects(start, spkcgraph["type"])
        ))

    data_type = get_settings_func(
        "VERIFIER_CLEAN_GRAPH",
        "spkcspider.apps.verifier.functions.clean_graph"
    )(mtype, g, start, source, hostpart)
    if not data_type:
        session.close()
        dvfile.close()
        os.unlink(dvfile.name)
        raise exceptions.ValidationError(
            _('Invalid graph type: %(type)s'),
            params={"type": data_type},
            code="invalid_type"
        )

    # retrieve further pages
    for page in range(2, pages+1):
        url = merge_get_url(
            source.get_url(), raw="embed", page=str(page)
        )
        # validation not neccessary here (base url is verified)
        params, can_inline = get_requests_params(url)
        if can_inline:
            resp = Client().get(url)
            if resp.status_code != 200:
                session.close()
                dvfile.close()
                os.unlink(dvfile.name)
                raise exceptions.ValidationError(
                    _("Retrieval failed: %(reason)s"),
                    params={"reason": resp.reason},
                    code="error_code:{}".format(resp.status_code)
                )

            c_length = resp.get("content-length", None)
            if not verify_download_size(c_length, current_size):
                resp.close()
                session.close()
                dvfile.close()
                os.unlink(dvfile.name)
                raise exceptions.ValidationError(
                    _("Content too big: %(size)s"),
                    params={"size": c_length},
                    code="invalid_size"
                )
            c_length = int(c_length)
            current_size += c_length
            # clear file
            dvfile.truncate(c_length)
            dvfile.seek(0, 0)

            for chunk in resp:
                dvfile.write(chunk)
            resp.close()
            dvfile.seek(0, 0)
        else:
            try:
                with session.get(
                    url, stream=True, **params
                ) as resp:
                    if resp.status_code != 200:
                        session.close()
                        dvfile.close()
                        os.unlink(dvfile.name)
                        raise exceptions.ValidationError(
                            _("Retrieval failed: %(reason)s"),
                            params={"reason": resp.reason},
                            code="error_code:{}".format(resp.status_code)
                        )

                    c_length = resp.headers.get("content-length", None)
                    if not verify_download_size(c_length, current_size):
                        session.close()
                        dvfile.close()
                        os.unlink(dvfile.name)
                        raise exceptions.ValidationError(
                            _("Content too big: %(size)s"),
                            params={"size": c_length},
                            code="invalid_size"
                        )
                    c_length = int(c_length)
                    current_size += c_length
                    # clear file
                    dvfile.truncate(c_length)
                    dvfile.seek(0, 0)

                    for chunk in resp.iter_content(BUFFER_SIZE):
                        dvfile.write(chunk)
                    dvfile.seek(0, 0)
            except requests.exceptions.Timeout:
                session.close()
                dvfile.close()
                os.unlink(dvfile.name)
                raise exceptions.ValidationError(
                    _('url timed out: %(url)s'),
                    params={"url": url},
                    code="timeout_url"
                )
            except requests.exceptions.ConnectionError:
                session.close()
                dvfile.close()
                os.unlink(dvfile.name)
                raise exceptions.ValidationError(
                    _('Invalid url: %(url)s'),
                    params={"url": url},
                    code="innvalid_url"
                )

        try:
            g.parse(
                dvfile.name,
                format="turtle"
            )
        except Exception as exc:
            if settings.DEBUG:
                dvfile.seek(0, 0)
                logging.error(dvfile.read())
            logging.exception(exc)
            session.close()
            dvfile.close()
            os.unlink(dvfile.name)
            # pages could have changed, but still incorrect
            raise exceptions.ValidationError(
                _("%(page)s is not a \"%(format)s\" file"),
                params={"format": "turtle", "page": page},
                code="invalid_file"
            )

        if task:
            task.update_state(
                state='RETRIEVING',
                meta={
                    'page': page,
                    'num_pages': pages
                }
            )

    hashable_nodes = set(g.subjects(
        predicate=spkcgraph["hashable"], object=Literal(True)
    ))

    hashes = [
        i for i in yield_hashes(g, hashable_nodes)
    ]
    if task:
        task.update_state(
            state='RETRIEVING',
            meta={
                'hashable_urls_checked': 0
            }
        )
    for count, t in enumerate(yield_hashable_urls(g, hashable_nodes), start=1):
        if (URIRef(t[2].value), None, None) in g:
            continue
        url = merge_get_url(t[2].value, raw="embed")
        if not get_settings_func(
            "SPIDER_URL_VALIDATOR",
            "spkcspider.apps.spider.functions.validate_url_default"
        )(url):
            session.close()
            dvfile.close()
            os.unlink(dvfile.name)
            raise exceptions.ValidationError(
                _('Insecure url: %(url)s'),
                params={"url": url},
                code="insecure_url"
            )

        params, can_inline = get_requests_params(url)
        if can_inline:
            resp = Client().get(url)
            if resp.status_code != 200:
                raise exceptions.ValidationError(
                    _("Retrieval failed: %(reason)s"),
                    params={"reason": resp.reason},
                    code="code_{}".format(resp.status_code)
                )
            h = get_hashob()
            h.update(XSD.base64Binary.encode("utf8"))
            for chunk in resp.iter_content(BUFFER_SIZE):
                h.update(chunk)
            resp.close()
            # do not use add as it could be corrupted by user
            # (user can provide arbitary data)
            g.set((
                URIRef(t[2].value),
                spkcgraph["hash"],
                Literal(h.finalize().hex())
            ))
            if task:
                task.update_state(
                    state='RETRIEVING',
                    meta={
                        'hashable_urls_checked': count
                    }
                )
        else:
            try:
                with session.get(
                    url, stream=True, **params
                ) as resp:
                    if resp.status_code != 200:
                        raise exceptions.ValidationError(
                            _("Retrieval failed: %(reason)s"),
                            params={"reason": resp.reason},
                            code="code_{}".format(resp.status_code)
                        )
                    h = get_hashob()
                    h.update(XSD.base64Binary.encode("utf8"))
                    for chunk in resp.iter_content(BUFFER_SIZE):
                        h.update(chunk)
                    # do not use add as it could be corrupted by user
                    # (user can provide arbitary data)
                    g.set((
                        URIRef(t[2].value),
                        spkcgraph["hash"],
                        Literal(h.finalize().hex())
                    ))
                    if task:
                        task.update_state(
                            state='RETRIEVING',
                            meta={
                                'hashable_urls_checked': count
                            }
                        )
            except requests.exceptions.Timeout:
                session.close()
                dvfile.close()
                os.unlink(dvfile.name)
                raise exceptions.ValidationError(
                    _('url timed out: %(url)s'),
                    params={"url": url},
                    code="timeout_url"
                )
            except requests.exceptions.ConnectionError:
                session.close()
                dvfile.close()
                os.unlink(dvfile.name)
                raise exceptions.ValidationError(
                    _('Invalid url: %(url)s'),
                    params={"url": url},
                    code="innvalid_url"
                )
    # not required anymore
    session.close()
    if task:
        task.update_state(
            state='HASHING',
        )

    # make sure triples are linked to start
    # (user can provide arbitary data)
    g.remove((start, spkcgraph["hashed"], None))
    for t in g.triples((None, spkcgraph["hash"], None)):
        g.add((
            start,
            spkcgraph["hashed"],
            t[0]
        ))
        hashes.append(binascii.unhexlify(t[2].value))

    for i in g.subjects(spkcgraph["type"], Literal("Content")):
        h = get_hashob()
        h.update(i.encode("utf8"))
        hashes.append(h.finalize())
    hashes.sort()

    h = get_hashob()
    for i in hashes:
        h.update(i)
    # do not use add as it could be corrupted by user
    # (user can provide arbitary data)
    digest = h.finalize().hex()
    g.set((
        start,
        spkcgraph["hash"],
        Literal(digest)
    ))

    dvfile.truncate(0)
    dvfile.seek(0, 0)
    # save in temporary file
    g.serialize(
        dvfile, format="turtle"
    )

    result, created = DataVerificationTag.objects.get_or_create(
        defaults={
            "dvfile": File(dvfile),
            "source": source,
            "data_type": data_type
        },
        hash=digest
    )
    dvfile.close()
    os.unlink(dvfile.name)
    update_fields = set()
    # and source, cannot remove source without replacement
    if not created and source and source != result.source:
        result.source = source
        update_fields.add("source")
    if data_type != result.data_type:
        result.data_type = data_type
        update_fields.add("data_type")
    result.save(update_fields=update_fields)
    if task:
        task.update_state(
            state='SUCCESS'
        )
    return result


@celery_app.task(bind=True, name='async validation')
def async_validate(self, ob, hostpart):
    ret = validate(ob, hostpart, self)
    return ret.get_absolute_url()


# def verify(tagid, task=None):
#    tag = DataVerificationTag.objects.get(id=tagid)
#    get_settings_func(
#        "VERIFIER_TAG_VERIFIER",
#        "spkcspider.apps.verifier.functions.verify_tag_default"
#    )(tag)
#    if tag.verification_state == "verified":
#        try:
#            tag.callback()
#        except exceptions.ValidationError:
#            logging.exception("Error while calling back")


# @celery_app.task(bind=True, name='async verification', ignore_results=True)
# def async_verify(self, tagid):
#     verify(tagid, self)



"""
Django settings for spkcspider project.

Generated by 'django-admin startproject' using Django 1.11.2.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

import os

import certifi
from django.utils.translation import gettext_lazy as _
from cryptography.hazmat.primitives import hashes

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
)


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.11/howto/deployment/checklist/


ALLOWED_HOSTS = []

FILE_UPLOAD_HANDLERS = [
    'django.core.files.uploadhandler.MemoryFileUploadHandler',
    "spkcspider.apps.spider.functions.LimitedTemporaryFileUploadHandler",
]


# Application definition

INSTALLED_APPS = [
    'widget_tweaks',
    'spkcspider.apps.spider_accounts',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',  # for flatpages
    'django.contrib.flatpages',
    'django.contrib.sitemaps',
    'spkcspider.apps.spider',
]
try:
    import django_extensions  # noqa: F401
    INSTALLED_APPS.append('django_extensions')
except ImportError:
    pass


MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.contrib.flatpages.middleware.FlatpageFallbackMiddleware',
]

ROOT_URLCONF = 'spkcspider.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'spkcspider.apps.spider.context_processors.settings',
                'django.template.context_processors.i18n',
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]
WSGI_APPLICATION = 'spkcspider.wsgi.application'

# Password validation
# https://docs.djangoproject.com/en/1.11/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',  # noqa: E501
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',  # noqa: E501
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',  # noqa: E501
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',  # noqa: E501
    },
]


# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = 'en'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


STATICFILES_DIRS = [
    # add node_modules as node_modules under static
    ("node_modules", "node_modules")
]

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/


STATIC_ROOT = 'static/'
STATIC_URL = '/static/'

MEDIA_ROOT = 'media/'
MEDIA_URL = '/media/'


CAPTCHA_CHALLENGE_FUNCT = 'captcha.helpers.math_challenge'
CAPTCHA_FONT_SIZE = 40

LOGIN_URL = "auth:login"
LOGIN_REDIRECT_URL = "auth:profile"
LOGOUT_REDIRECT_URL = "home"

AUTH_USER_MODEL = 'spider_accounts.SpiderUser'
# uses cryptography, can automatically determinate size
SPIDER_HASH_ALGORITHM = hashes.SHA512()
SPIDER_MIN_STRENGTH_EVELATION = 2
# change size of request token.
# Note: should be high to prevent token exhaustion
# TOKEN_SIZE = 30
# Change size of token for files (should be >=TOKEN_SIZE),
#  defaults to TOKEN_SIZE
# FILE_TOKEN_SIZE
# OPEN_FOR_REGISTRATION = True # allow registration
# ALLOW_USERNAME_CHANGE = True # allow users changing their username

## Default static token size (!=TOKEN_SIZE) # noqa: E266
# SPIDER_INITIAL_STATIC_TOKEN_SIZE

## captcha field names (REQUIRED) # noqa: E266
SPIDER_CAPTCHA_FIELD_NAME = "sunglasses"

## Update dynamic content, ... after migrations, default=true  # noqa: E266
# disable when importing backup
# ease deploy
UPDATE_DYNAMIC_AFTER_MIGRATION = True

# controls inlining of requests calls (default: None)
# can take a function with specification func(url) -> bool
#
#   by faking requests calls on urls which are internal
#   deadlocking hazards are avoided and the speed is increased
#   so it is really recommended to not set this setting on False
# SPIDER_INLINE = False

## extensions of images (used in file_filets)  # noqa: E266
# SPIDER_IMAGE_EXTENSIONS
## extensions of media (used in file_filets)  # noqa: E266
# SPIDER_MEDIA_EXTENSIONS

## embeddding function for files in graph, for e.g. linking  # noqa: E266
# SPIDER_FILE_EMBED_FUNC

## validator function for url requests  # noqa: E266
# SPIDER_URL_VALIDATOR

## Enable captchas  # noqa: E266
# INSTALLED_APPS.append('captcha')
# USE_CAPTCHAS = True

## Approval function for allowing dangerous self and travel protections  # noqa: E266, E501
##   Note: return False means don't allow and None ask admin  # noqa: E266
# SPIDER_DANGEROUS_APPROVE

# DIRECT_FILE_DOWNLOAD = True

# SPIDER_CONTENTVARIANT_FILTER

# SPIDER_TAG_VERIFIER_VALIDATOR
# SPIDER_TAG_VERIFY_REQUEST_VALIDATOR

# SPIDER_ANCHOR_DOMAIN
# SPIDER_COMPONENTS_DELETION_PERIODS
# SPIDER_CONTENTS_DEFAULT_DELETION_PERIOD
# SPIDER_RATELIMIT_FUNC
## Enable direct file downloads (handled by webserver)  # noqa: E266
# disadvantage: blocking access requires file name change
# FILE_DIRECT_DOWNLOAD
# FILE_FILET_DIR
# FILE_FILET_SALT_SIZE
# SPIDER_UPLOAD_FILTER
# SPIDER_GET_QUOTA
# SPIDER_MAX_FILE_SIZE
# SPIDER_MAX_FILE_SIZE_STAFF
# SPIDER_USER_QUOTA_LOCAL
# SPIDER_USER_QUOTA_REMOTE
## in units  # noqa: E266
# SPIDER_USER_QUOTA_USERCOMPONENTS

# unbreak old links after switch to a new machine friendly url layout
SPIDER_LEGACY_REDIRECT = True

##  Use subpath to create ids for identifiers # noqa: E266
# SPIDER_ID_USE_SUBPATH

# usercomponents created with user
SPIDER_DEFAULT_COMPONENTS = {
    "home": {
        "public": False,
        "features": ["Persistence", "WebConfig"]
    },
    "public": {
        "public": True,
        "features": []
    },
}

## Default description  # noqa: E266
SPIDER_DESCRIPTION = "A spkcspider instance for my personal data."

SPIDER_BLACKLISTED_MODULES = []

# maximal domain_mode activation per usercomponent/domain
SPIDER_DOMAIN_UPDATE_RATE = "10/m"
# maximal error rate for a domain before blocking requests
SPIDER_DOMAIN_ERROR_RATE = "10/10m"
# max description length (stripped)
SPIDER_MAX_DESCRIPTION_LENGTH = 200
# how many user components/contents per page
SPIDER_OBJECTS_PER_PAGE = 25
# how many raw/serialized results per page?
SPIDER_SERIALIZED_PER_PAGE = 50
# max depth of references
SPIDER_MAX_EMBED_DEPTH = 5
# how many search parameters are allowed
SPIDER_MAX_SEARCH_PARAMETERS = 30
# licences for media
SPIDER_LICENSE_CHOICES = {
    "other": {
        "url": ""
    },
    "pd": {
        "name": _("Public Domain/CC0"),
        "url":
            "https://creativecommons.org/publicdomain/zero/1.0/legalcode"
    },
    "CC BY": {
        "url": "https://creativecommons.org/licenses/by/4.0/legalcode"
    },
    "CC BY-SA": {
        "url": "https://creativecommons.org/licenses/by-sa/4.0/legalcode"
    },
    "CC BY-ND": {
        "url": "https://creativecommons.org/licenses/by-nd/4.0/legalcode"
    },
    "CC BY-NC": {
        "url": "https://creativecommons.org/licenses/by-nc/4.0/legalcode"
    },
    "CC BY-NC-SA": {
        "url": "https://creativecommons.org/licenses/by-nc-sa/4.0/legalcode"
    },
    "CC BY-NC-ND": {
        "url": "https://creativecommons.org/licenses/by-nc-nd/4.0/legalcode"
    },
}
SPIDER_DEFAULT_LICENSE_FILE = "CC BY"

# licences for text (default: file licenses are used)
# SPIDER_LICENSE_CHOICES_TEXT
# SPIDER_DEFAULT_LICENSE_TEXT

# requests parameter overwrites
# * "hostname.foo": parameter for specific domain
# * "".foo": parameter for a tld
# * b"default": default parameters for request
# why binary? Because it cannot clash with a "default" host this way
# hierarchy: host > tld > b"default"
SPIDER_REQUEST_KWARGS_MAP = {
    b"default": {
        "verify": certifi.where(),
        "timeout": 3,
        "proxies": {}
    },
    # example for usage with tor (requires requests[socks])
    # ".onion": {
    #     "timeout": 10,
    #     "proxies": {
    #        'http': 'socks5://localhost:9050',
    #        'https': 'socks5://localhost:9050'
    #     }
    # }
    # example for a slow domain
    # "veryslow.example": {
    #     "timeout": 60,
    #     "proxies": {}
    # }
}


# for sites
SITE_ID = 1




__all__ = (
    "ProtectionType", "VariantType", "ProtectionResult",
    "MIN_PROTECTION_STRENGTH_LOGIN",
    "TravelLoginType", "MAX_TOKEN_SIZE", "MAX_TOKEN_B64_SIZE",
    "hex_size_of_bigid", "TokenCreationError", "protected_names", "spkcgraph",
    "dangerous_login_choices", "ActionUrl", "static_token_matcher",
    "host_tld_matcher", "travel_scrypt_params", "loggedin_active_tprotections",
    "essential_contents"
)

import enum
from collections import namedtuple
import re

from rdflib.namespace import Namespace


spkcgraph = Namespace("https://spkcspider.net/static/schemes/spkcgraph#")
# Literal allows arbitary datatypes, use this and don't bind

hex_size_of_bigid = 16

MAX_TOKEN_SIZE = 90

if MAX_TOKEN_SIZE % 3 != 0:
    raise Exception("MAX_TOKEN_SIZE must be multiple of 3")

MAX_TOKEN_B64_SIZE = MAX_TOKEN_SIZE*4//3


MIN_PROTECTION_STRENGTH_LOGIN = 2


# user can change static token but elsewise the token stays static
static_token_matcher = re.compile(
    r"(?:[^?]*/|^)(?P<static_token>[^/?]+)/(?P<access>[^/?]+)"
)

host_tld_matcher = re.compile(
    r'^[^.]*?(?!\.)(?P<host>[^/?:]+?(?P<tld>\.[^/?:.]+)?)(?=[/?:]|$)(?!:/)'
)


essential_contents = {
    "domainmode", "defaultactions"
}


class TokenCreationError(Exception):
    pass


ProtectionResult = namedtuple("ProtectionResult", ["result", "protection"])
ActionUrl = namedtuple("ActionUrl", ["name", "url"])
protected_names = {"index"}


travel_scrypt_params = {
    "length": 32,
    "n": 2**14,
    "r": 16,
    "p": 2
}


class ProtectionType(str, enum.Enum):
    # receives: request, scope
    access_control = "a"
    # receives: request, scope, password
    authentication = "b"
    # c: former reliable, no meaning
    # protections which does not contribute to required_passes (404 only)
    no_count = "d"
    # protections which have side effects
    side_effects = "e"
    # forget about recovery, every recovery method is authentication
    # and will be misused this way
    # The only real recovery is by staff and only if there is a secret


class VariantType(str, enum.Enum):
    # a not assigned
    # use persistent token
    #  can be used for some kind of federation
    #  warning: if token is deleted persisted content is deleted
    persist = "b"
    # update content is without form/for form updates it is not rendered
    # required for still beeing able to update elemental parameters
    raw_update = "c"
    # raw_add not required, archieved by returning response

    # don't list as contentvariant for user (for computer only stuff)
    # set unlisted as default in info field
    # if feature: unlist as a public visible feature
    unlisted = "d"
    # activates domain mode
    domain_mode = "e"
    # allow outside applications push content to spiders
    # adds by default unlisted info attribute
    # appears in features of userComponent
    # counts as foreign content
    # don't list as content variant for user (
    #    except feature_connect and not unique
    # )
    component_feature = "f"
    # same, but assigns to contents
    content_feature = "g"

    # is content unique for usercomponent
    # together with strength level 10: unique for user
    unique = "h"
    # can be used as anchor, will be hashed on verification if primary
    anchor = "i"
    # unique:
    #  auto create if content does not exist, auto unset if content is deleted
    # non-unique:
    #  add as contentvariant (except if unlisted), and create feature
    feature_connect = "j"
    # add as machine creatable content
    machine = "k"
    # exclude from exports
    no_export = "l"


class TravelLoginType(str, enum.Enum):
    # don't show protected contents and components
    hide = "a"
    # switches to hiding if trigger was activated
    trigger_hide = "b"
    # disable login
    disable = "c"
    # disable login if triggered
    trigger_disable = "C"
    # wipe travel protected contents and components
    # Note: noticable if shared contents are removed
    wipe = "d"
    # wipe user, self destruct user on login, Note: maybe noticable
    wipe_user = "e"


dangerous_login_choices = {
    TravelLoginType.wipe.value,
    TravelLoginType.wipe_user.value
}

loggedin_active_tprotections = {
    TravelLoginType.hide.value,
    TravelLoginType.disable.value
}



# flake8: noqa

from spkcspider.settings import *  # noqa: F403, F401

INSTALLED_APPS += [
    'spkcspider.apps.spider_filets',
    'spkcspider.apps.spider_keys',
    'spkcspider.apps.spider_tags',
    'spkcspider.apps.spider_webcfg',
    # ONLY for tests and REAL verifiers=companies verifing data
    'spkcspider.apps.verifier',
    'captcha'
]
USE_CAPTCHAS = True


# for tests
SPIDER_REQUEST_KWARGS_MAP["localhost"] = {
    "verify": False,
    "timeout": 1,
    "proxies": {}
}

# Verifier specific options, normally not required
# requests parameter overwrites
# * "hostname.foo": parameter for specific domain
# * "".foo": parameter for a tld
# * b"default": default parameters for request
# why binary? Because it cannot clash with a "default" host this way
# hierarchy: host > tld > b"default"
# NOTE: if VERIFIER_REQUEST_KWARGS_MAP is not specified
#       SPIDER_REQUEST_KWARGS_MAP is used
VERIFIER_REQUEST_KWARGS_MAP = {
    b"default": {
        "verify": certifi.where(),
        "timeout": 6,
        "proxies": {}
    },
    "localhost": {  # for tests
        "verify": False,
        "timeout": 3,
        "proxies": {}
    }
}
# how many verification requests of user/ip per minute
VERIFIER_REQUEST_RATE = "10/m"
# 40 mb maximal size
VERIFIER_MAX_SIZE_ACCEPTED = 40000000
# 2 mb, set to 0 to disable a direct file upload
VERIFIER_MAX_SIZE_DIRECT_ACCEPTED = 2000000
# controls inlining of verifier, defaults to SPIDER_INLINE
# VERIFIER_INLINE = False

# not required, SpiderTokenAuthBackend have to be tested, so here active
AUTHENTICATION_BACKENDS = [
    'django.contrib.auth.backends.ModelBackend',
    'spkcspider.apps.spider.auth.SpiderTokenAuthBackend'
]

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '^_08u&*be(*my6$pv^m3fki!2s5)5e)9@l5l#srch1h)w3p+$l'

# Database
# https://docs.djangoproject.com/en/1.11/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

# how many user contents/components per page
SPIDER_OBJECTS_PER_PAGE = 3
# how many raw/serialized results per page?
SPIDER_SERIALIZED_PER_PAGE = 3
# max depth of references
SPIDER_MAX_EMBED_DEPTH = 5

CELERY_BROKER_URL = "redis://127.0.0.1:6379"
CELERY_RESULT_BACKEND = "redis://127.0.0.1:6379"

# specify fixtures directory for tests
FIXTURE_DIRS = [
    "tests/fixtures/"
]



__all__ = ["SpiderVerifierConfig"]

from django.apps import AppConfig


class SpiderVerifierConfig(AppConfig):
    name = 'spkcspider.apps.verifier'
    label = 'spider_verifier'
    verbose_name = 'spkcspider verifier'
    spider_url_path = 'verify/'



from django import forms
from django.contrib.auth.forms import UserCreationForm, UserChangeForm
from django.contrib.auth import get_user_model
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.urls import reverse
from django.conf import settings


if getattr(settings, "USE_CAPTCHAS", False):
    from captcha.fields import CaptchaField
else:
    CaptchaField = None


class SignupForm(UserCreationForm):
    # TODO: for a better spammer protection, use useragent for seeding
    # and add/rename, move random fields
    # disguise hidden fields
    use_required_attribute = False

    # real: username, fake for simple spammers
    name = forms.CharField(
        label="", max_length=100, required=False, widget=forms.HiddenInput()
    )
    # fake for simple spammers
    email = forms.EmailField(
        label="", max_length=100, required=False, widget=forms.HiddenInput()
    )
    # email, real
    liame = forms.EmailField(
        label=_('Email Address'), max_length=100, required=False,
        help_text=_("(optional)")
    )

    class Meta:
        model = get_user_model()
        fields = ['username', 'password1', 'password2',
                  'name', 'liame', 'email']

    def is_valid(self):
        """
        Returns True if the form has no errors. Otherwise, False. If errors are
        being ignored, returns False.
        """
        if self.data['email'] or self.data['name']:
            return False
        return self.is_bound and not self.errors

    def clean(self):
        # clean up email hack
        super(SignupForm, self).clean()
        self.cleaned_data["email"] = self.cleaned_data["liame"]
        del self.cleaned_data["liame"]
        del self.cleaned_data["name"]


if CaptchaField:
    # add captcha, disguised name
    SignupForm.declared_fields[settings.SPIDER_CAPTCHA_FIELD_NAME] = \
        CaptchaField(label=_("Captcha"))
    SignupForm.base_fields[settings.SPIDER_CAPTCHA_FIELD_NAME] = \
        SignupForm.declared_fields[settings.SPIDER_CAPTCHA_FIELD_NAME]


class UserUpdateForm(UserChangeForm):
    class Meta(UserChangeForm.Meta):
        model = get_user_model()
        fields = model.SAFE_FIELDS

    def __init__(self, *args, **kwargs):
        _ = gettext
        super(UserChangeForm, self).__init__(*args, **kwargs)
        self.fields['password'].help_text = \
            self.fields['password'].help_text.format(
                reverse('auth:password_change')
            )
        f = self.fields.get('user_permissions')
        if f is not None:
            f.queryset = f.queryset.select_related('content_type')
        self.fields['email'].help_text += _("Optional")



# Generated by Django 2.1.3 on 2018-11-21 07:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0007_auto_20181016_2002'),
    ]

    operations = [
        migrations.AlterField(
            model_name='textfilet',
            name='text',
            field=models.TextField(blank=True, default=''),
        ),
    ]



__all__ = ["SpiderAccountsConfig"]

from django.apps import AppConfig


class SpiderAccountsConfig(AppConfig):
    name = 'spkcspider.apps.spider_accounts'
    label = 'spider_accounts'
    verbose_name = 'Spkcspider User Model'
    spider_url_path = "accounts/"
